/*
 * YouGrabber
 *
 * A lightweight, ncurses-based, multi-threaded command line YouTube.com
 * video downloader with proxy support, made from scratch in ANSI C.
 *
 * Copyright (C) 2006-2007 Quetzy Garcia, <quetzyg@users.sourceforge.net>
 * For news and updates, see <http://yougrabber.sourceforge.net/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "common.h"

static const double __MEGABYTE = 1024.0*1024.0;
static const double __KILOBYTE = 1024.0;
static const double __MINUTE = 60.0;
static const double __PERCENT = 100.0;

/* fixed terminal sizes */
static const unsigned short __LINES = 24;
static const unsigned short __COLS = 80;

static const unsigned short __FRAME_HEIGHT = 10;

/* interface pointer */
static I *__i_ptr = NULL;

/* current frame pointer */
static F *__f_ptr = NULL;

struct __frame
{
  WINDOW *win_box;
  WINDOW *win_sub;
  PANEL *panel;
};

struct __interface
{
  C *c_ptr;
  GList *f_list;
};

/* clear window lines */
static void __interface_clrwlines(WINDOW *win, int beg_line, int num_lines)
{
  int i;

  for (i = beg_line; i < (beg_line+num_lines); ++i)
  {
    wmove(win, i, 0);
    wclrtoeol(win);
  }
}

/* start ncurses */
void interface_nc_start(void)
{
  initscr();
  cbreak();
  noecho();
  curs_set(0);
  keypad(stdscr, TRUE);
}

/* stop ncurses */
void interface_nc_stop(void)
{
  clear();
  refresh();
  endwin();
}

/* create a new frame and return pointer */
F *interface_frame_new(D *d_ptr)
{
  F *f_ptr = NULL;

  if (d_ptr == NULL)
    return NULL;

  if (!(f_ptr = (F*) malloc(sizeof(F))))
    return NULL;

  download_set_user_pointer(d_ptr, f_ptr);

  f_ptr->win_box = newwin(__FRAME_HEIGHT, __COLS, __LINES-__FRAME_HEIGHT-1, 0);
  box(f_ptr->win_box, 0, 0);

  f_ptr->win_sub = derwin(f_ptr->win_box, __FRAME_HEIGHT-2, __COLS-2, 1, 1);

  f_ptr->panel = new_panel(f_ptr->win_box);
  set_panel_userptr(f_ptr->panel, d_ptr);

  return f_ptr;
}

/* free frame resources (GFunc() type function) */
void interface_frame_free(gpointer data, gpointer user_data)
{
  (void) user_data;
  F *f_ptr = NULL;

  f_ptr = (F*) data;

  download_free(panel_userptr(f_ptr->panel));
  del_panel(f_ptr->panel);
  delwin(f_ptr->win_sub);
  delwin(f_ptr->win_box);
  free(f_ptr);
}

/* write to main interface window */
void interface_write_main(C *c_ptr)
{
  char *login_status[] =
  {
    "failed (flagged downloads will be unavailable)",
    "skipped (flagged downloads will be unavailable)",
    "waiting...",
    "logged"
  };

  char buf[128];
  char *file_path = NULL;
  char *proxy_host = NULL;
  long proxy_port;

  file_path = config_get_file_path(c_ptr);
  proxy_host = config_get_proxy_host(c_ptr);
  proxy_port = config_get_proxy_port(c_ptr);

  __interface_clrwlines(stdscr, 0, 9);

  mvwprintw(stdscr, 0, 0, "%s v%s - Copyright (C) %s <%s>",
    YG_STRING, YG_VERSION, YG_AUTHOR, YG_EMAIL);

  mvwhline(stdscr, 1, 0, ACS_HLINE, __COLS);

  if (strlen(file_path) > 0)
    g_snprintf(buf, sizeof(buf), "'%s'", file_path);
  else
    g_strlcpy(buf, "default", sizeof(buf));
  mvwprintw(stdscr, 3, 1, "SETTINGS: %s", buf);

  if (strlen(proxy_host) > 0)
    g_snprintf(buf, sizeof(buf), "via proxy (%s:%ld)", proxy_host, proxy_port);
  else
    g_strlcpy(buf, "direct", sizeof(buf));
  mvwprintw(stdscr, 4, 1,  "CONNECTION: %s", buf);

  mvwprintw(stdscr, 5, 1, "YOUTUBE LOGIN STATE: %s",
    login_status[config_get_youtube_login_status(c_ptr)]);

  mvwprintw(stdscr, 8, 1, "DOWNLOADS REMAINING: %.3d", dip_get());

  /* footer */
  __interface_clrwlines(stdscr, __LINES-1, 1);

  if (dip_get() > 0)
    g_strlcpy(buf, "Wait for downloads to finish", sizeof(buf));
  else
    g_strlcpy(buf, "Press ESC to exit", sizeof(buf));

  mvwprintw(stdscr, __LINES-1, 0,
    "LEFT/RIGHT ARROW KEYS: browse between downloads - %s", buf);
}

/* write to current frame window */
void interface_write_frame(F *f_ptr)
{
  D *d_ptr = NULL;
  double dlnow = 0.0;
  double dltotal = 0.0;
  double time = 0.0;
  char *url_ptr = NULL;

  char *download_status[] =
  {
    "video is unavailable",
    "video has been removed",
    "I/O error",
    "network error", /* the proper error message will be shown */
    "downloading video data...",
    "download finished"
  };

  if ((d_ptr = panel_userptr(f_ptr->panel)))
  {
    dlnow = download_get_file_size_now(d_ptr);
    dltotal = download_get_file_size_total(d_ptr);
    time = download_get_time(d_ptr);
    url_ptr = download_get_video_url(d_ptr);

    /* download box title */
    mvwprintw(f_ptr->win_box, 0, 2, " DOWNLOAD %.3d/%.3d ",
      g_list_index(__i_ptr->f_list, f_ptr)+1, g_list_length(__i_ptr->f_list));

    /* download box body info */
    __interface_clrwlines(f_ptr->win_sub, 0, 9);

    mvwprintw(f_ptr->win_sub, 0, 0, "TITLE: %s",
      download_get_video_title(d_ptr));

    mvwprintw(f_ptr->win_sub, 1, 0, "FILE: %s",
      download_get_file_name(d_ptr));

    if (download_get_status(d_ptr) == DOWNLOAD_ERROR_NETWORK)
      mvwprintw(f_ptr->win_sub, 2, 0, "STATUS: %s",
        download_get_network_error(d_ptr));
    else
      mvwprintw(f_ptr->win_sub, 2, 0, "STATUS: %s",
        download_status[download_get_status(d_ptr)]);

    mvwprintw(f_ptr->win_sub, 3, 0, "DOWNLOADED: %.2f%% [%.2f of %.2f MB]\n",
      (dlnow*__PERCENT)/dltotal, dlnow/__MEGABYTE, dltotal/__MEGABYTE);

    mvwprintw(f_ptr->win_sub, 4, 0, "SPEED: %.2f KB/sec",
      download_get_speed(d_ptr)/__KILOBYTE);

    mvwprintw(f_ptr->win_sub, 5, 0, "ELAPSED: %d min %.f sec",
      (int) (time/__MINUTE), fmod(time, __MINUTE));

    /* truncate long URLs */
    if (strlen(url_ptr) > 128)
      mvwprintw(f_ptr->win_sub, 6, 0, "URL: %.128s... (URL truncated)", url_ptr);
    else
      mvwprintw(f_ptr->win_sub, 6, 0, "URL: %s", url_ptr);
  }
}

/* refresh all interface windows */
void interface_refresh()
{
  wnoutrefresh(stdscr);
  wnoutrefresh(__f_ptr->win_box);
  wnoutrefresh(__f_ptr->win_sub);
  doupdate();
}

/* create new interface and return pointer */
I *interface_new(C *c_ptr, GList *f_list)
{
  if (!(__i_ptr = (I*) malloc(sizeof(I))))
    return NULL;

  __i_ptr->c_ptr = c_ptr;
  __i_ptr->f_list = f_list;

  return __i_ptr;
}

/* free interface resources */
void interface_free(I *i_ptr)
{
  g_list_foreach(i_ptr->f_list, interface_frame_free, NULL);
  g_list_free(i_ptr->f_list);

  config_free(i_ptr->c_ptr);

  free(i_ptr);
}

/* display interface */
int interface_main_loop(I *i_ptr)
{
  unsigned quit = 0;
  GList *node_current = NULL;
  GList *node_next = NULL;

  node_current = g_list_first(i_ptr->f_list);

  while(!quit)
  {
    node_current = (!node_next) ? node_current : node_next;
    __f_ptr = node_current->data;

    interface_write_main(i_ptr->c_ptr);
    interface_write_frame(__f_ptr);
    interface_refresh();

    top_panel(__f_ptr->panel);
    update_panels();
    doupdate();

    switch (getch())
    {
      case KEY_LEFT:
        node_next = g_list_previous(node_current);
      break;

      case KEY_RIGHT:
        node_next = g_list_next(node_current);
      break;

      case 27: /* ESC */
        if (dip_get() == 0)
          quit = 1;
      break;
    }
  }

  return 0;
}
